"""DatasetMetadata <-> proto converters."""

from __future__ import annotations

from typing import TYPE_CHECKING

from radiens_core._converters._channels import channel_metadata_from_proto
from radiens_core._generated import common_pb2
from radiens_core.models.dataset import DatasetMetadata, RadiensFileType
from radiens_core.models.time_range import TimeRange

if TYPE_CHECKING:
    from radiens_core._generated import datasource_pb2


# --- RadiensFileType enum conversion ---

_FILE_TYPE_FROM_PROTO: dict[int, RadiensFileType] = {
    int(common_pb2.RHD): RadiensFileType.RHD,
    int(common_pb2.XDAT): RadiensFileType.XDAT,
    int(common_pb2.CSV): RadiensFileType.CSV,
    int(common_pb2.HDF5): RadiensFileType.HDF5,
    int(common_pb2.NEX5): RadiensFileType.NEX5,
    int(common_pb2.NWB): RadiensFileType.NWB,
    int(common_pb2.KILOSORT2): RadiensFileType.KILOSORT2,
    int(common_pb2.NSX): RadiensFileType.NSX,
    int(common_pb2.TDT): RadiensFileType.TDT,
    int(common_pb2.SPIKES): RadiensFileType.SPIKES,
    int(common_pb2.OE_DAT): RadiensFileType.OE_DAT,
}

_FILE_TYPE_TO_PROTO: dict[RadiensFileType, common_pb2.RadixFileTypes.ValueType] = {
    RadiensFileType.RHD: common_pb2.RHD,
    RadiensFileType.XDAT: common_pb2.XDAT,
    RadiensFileType.CSV: common_pb2.CSV,
    RadiensFileType.HDF5: common_pb2.HDF5,
    RadiensFileType.NEX5: common_pb2.NEX5,
    RadiensFileType.NWB: common_pb2.NWB,
    RadiensFileType.KILOSORT2: common_pb2.KILOSORT2,
    RadiensFileType.NSX: common_pb2.NSX,
    RadiensFileType.TDT: common_pb2.TDT,
    RadiensFileType.SPIKES: common_pb2.SPIKES,
    RadiensFileType.OE_DAT: common_pb2.OE_DAT,
}


def file_type_from_proto(pb: int) -> RadiensFileType:
    """Convert proto RadixFileTypes int to domain RadiensFileType."""
    return _FILE_TYPE_FROM_PROTO[pb]


def file_type_to_proto(ft: RadiensFileType) -> common_pb2.RadixFileTypes.ValueType:
    """Convert domain RadiensFileType to proto RadixFileTypes."""
    return _FILE_TYPE_TO_PROTO[ft]


def guess_file_type(path: str) -> RadiensFileType:
    """Guess file type from file extension."""
    ext = path.rsplit(".", maxsplit=1)[-1].lower() if "." in path else ""
    ext_map: dict[str, RadiensFileType] = {
        "xdat": RadiensFileType.XDAT,
        "rhd": RadiensFileType.RHD,
        "csv": RadiensFileType.CSV,
        "hdf5": RadiensFileType.HDF5,
        "h5": RadiensFileType.HDF5,
        "nex5": RadiensFileType.NEX5,
        "nwb": RadiensFileType.NWB,
        "bin": RadiensFileType.KILOSORT2,
        "nsx": RadiensFileType.NSX,
        "tdt": RadiensFileType.TDT,
        "oebin": RadiensFileType.OE_DAT,
    }
    return ext_map.get(ext, RadiensFileType.XDAT)


def dataset_metadata_from_proto(
    raw: datasource_pb2.DataSourceSetSaveReply,
) -> DatasetMetadata:
    """Build domain DatasetMetadata from proto DataSourceSetSaveReply."""
    status = raw.status
    sig_group = status.signalGroup
    pb_tr = status.tR

    channel_metadata = channel_metadata_from_proto(sig_group)

    # Build TimeRange from the proto TimeRange message
    sec_list = list(pb_tr.timeRangeSec)
    ts_list = list(pb_tr.timestamp)
    sec_start = sec_list[0] if len(sec_list) > 0 else 0.0
    sec_end = sec_list[1] if len(sec_list) > 1 else 0.0
    ts_start = ts_list[0] if len(ts_list) > 0 else 0
    ts_end = ts_list[1] if len(ts_list) > 1 else 0

    time_range = TimeRange(
        sec=(sec_start, sec_end),
        timestamp=(ts_start, ts_end),
        fs=pb_tr.fs,
        dur_sec=sec_end - sec_start,
        walltime_str=pb_tr.wallTimeStart,
        n=ts_end - ts_start,
    )

    return DatasetMetadata(
        id=raw.dsourceID,
        base_name=status.baseFileName,
        path=status.path,
        file_type=file_type_from_proto(status.fileType),
        time_range=time_range,
        channel_metadata=channel_metadata,
        parent_dsource_id="",
        probe_uid=sig_group.datasetUID,
    )



"""VidereClient — offline file-based analysis client."""

from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, cast

from pydantic import ConfigDict, validate_call

from radiens_core._base_client import BaseClient
from radiens_core._converters._signals import sig_select_to_proto, signal_arrays_from_signals2
from radiens_core._converters._time_range import time_range_to_proto, trs_mode_to_proto
from radiens_core._converters._validation import validate_proto
from radiens_core._generated import common_pb2
from radiens_core._services._radiens_core import get_hd_snapshot_py_proto, set_datasource
from radiens_core._services._radiens_lifecycle import healthcheck_proto
from radiens_core._transport._channel_pool import ChannelPool
from radiens_core._transport._discovery import discover_radiens_addresses
from radiens_core.exceptions import RadiensError
from radiens_core.models.common import ClientType, ServiceEndpoint
from radiens_core.models.dataset import DatasetMetadata, FlexTimeRange
from radiens_core.models.signals import FlexSignalSpec, SignalArrays

if TYPE_CHECKING:
    from radiens_core.models.dataset import TimeRangeSpec
    from radiens_core.models.signals import SignalSpec
    from radiens_core.models.types import DatasetRef


class VidereClient(BaseClient):
    """Client for offline file-based analysis.

    Auto-discovers a running ``radiensserver`` on the local machine.

    Usage::

        with VidereClient() as client:
            meta = client.link_data_file("/path/to/recording.xdat")
            sigs = client.get_signals(meta, time_range=[0, 1], signals=[0, 1, 2])
    """

    def __init__(self) -> None:
        # 1. Discover addresses
        tmp_pool = ChannelPool()
        try:
            addresses = discover_radiens_addresses(tmp_pool)
        finally:
            tmp_pool.close()

        # 2. Base init
        core_addr = addresses.get(ServiceEndpoint.RADIENS_CORE, "")
        super().__init__(
            addresses=addresses,
            auth_server_address=core_addr,
            fallback_service=ServiceEndpoint.RADIENS_CORE,
        )

        self._datasets: dict[str, DatasetMetadata] = {}

    def _resolve_dataset(self, ref: DatasetRef) -> DatasetMetadata:
        """Resolve a dataset reference to DatasetMetadata."""
        if isinstance(ref, DatasetMetadata):
            return ref
        ref_str = str(ref)
        if ref_str in self._datasets:
            return self._datasets[ref_str]
        raise ValueError(f"Unknown dataset: {ref_str}. Link it first with link_data_file().")

    # --- Public API ---

    def healthcheck(self, service: ServiceEndpoint = ServiceEndpoint.RADIENS_LIFECYCLE) -> None:
        """Verify the radiensserver is responsive.

        Parameters
        ----------
        service
            Which gRPC service to probe (default ``RADIENS_LIFECYCLE``).
            ``RADIENS_CORE`` is also supported.

        Raises
        ------
        RadiensError
            If the health check fails or the server is unresponsive.
        """

        addr = self._server_address(service)
        response = healthcheck_proto(self._pool, addr, ClientType.VIDERE, service=service)
        if response.errMsg:
            raise RadiensError(f"Health check failed: {response.errMsg}")

    def link_data_file(
        self,
        path: str | Path,
    ) -> DatasetMetadata:
        """Link a recording file and return its metadata.

        Parameters
        ----------
        path
            Path to the recording file (e.g., .xdat, .rhd).

        Returns
        -------
        DatasetMetadata
            Metadata for the linked dataset.
        """

        addr = self._server_address(ServiceEndpoint.RADIENS_CORE)
        meta = set_datasource(self._pool, addr, str(path))
        self._datasets[meta.id] = meta
        return meta

    @validate_call(config=ConfigDict(arbitrary_types_allowed=True))
    def get_signals(
        self,
        dataset: str | Path | DatasetMetadata,
        time_range: FlexTimeRange,
        signals: FlexSignalSpec,
    ) -> SignalArrays:
        """Retrieve signal data for a specific time range and channel selection.

        Parameters
        ----------
        dataset
            Dataset to query (metadata object or ID).
        time_range
            Time range specification (e.g., [start, end] in seconds).
        signals
            Signal selection (e.g., list of amplifier channel indices).

        Returns
        -------
        SignalArrays
            The requested signal data as numpy arrays.
        """

        meta = self._resolve_dataset(dataset)

        # Coerce flex types to specs (redundant due to @validate_call but satisfies mypy)
        tr_spec = cast("TimeRangeSpec", time_range)
        sig_spec = cast("SignalSpec", signals)

        # Resolve specs into concrete domain models
        concrete_tr = tr_spec.resolve(dataset_metadata=meta)
        if concrete_tr is None:
            # FROM_HEAD (lookback) mode is not meaningful for offline files
            raise ValueError(
                f"lookback() time range is not supported for offline files; "
                f"use TimeRangeSpec.subset(start, end) or TimeRangeSpec.full() instead. Got: {time_range}"
            )

        sig_select = sig_spec.to_sig_select(meta.channel_metadata)

        # Build proto request
        req = common_pb2.HDSnapshotRequest2(
            dsourceID=meta.id,
            tR=time_range_to_proto(concrete_tr),
            selMode=trs_mode_to_proto(tr_spec.mode),
            sigSel=sig_select_to_proto(sig_select),
        )
        validate_proto(req)

        # Execute RPC
        addr = self._server_address(ServiceEndpoint.RADIENS_CORE)
        response = get_hd_snapshot_py_proto(self._pool, addr, req)

        return signal_arrays_from_signals2(response.sigs)



"""Tests for AllegoClient.get_signals."""

from __future__ import annotations

from unittest.mock import patch

import numpy as np

from radiens_core._constants import PRIMARY_CACHE_STREAM_GROUP_ID
from radiens_core._generated import common_pb2
from radiens_core.allego_client import AllegoClient
from radiens_core.models.channels import ChannelMetadata, KeyIdxs, SignalUnits
from radiens_core.models.dataset import TimeRangeSpec
from radiens_core.models.signals import SignalArrays, SignalType


def _make_channel_metadata(fs: float = 30000.0) -> ChannelMetadata:
    sig_map = {st: KeyIdxs(ntv=[0], dset=[0], sys=[0]) for st in SignalType}
    return ChannelMetadata(
        dataset_uid="live",
        fs=fs,
        source_label="test_source",
        neighbor_max_radius_um=100.0,
        signal_units={SignalType.AMP: SignalUnits.MICROVOLTS},
        has_discrete=False,
        has_continuous_amp=True,
        sig_map=sig_map,
        selected_sig_map=sig_map,
        num_channels={st: 1 for st in SignalType},
        selected_num_channels={st: 1 for st in SignalType},
        site_positions=[],
    )


def _make_hd_snapshot(n_channels: int = 1, n_samples: int = 30000) -> common_pb2.HDSnapshot:
    data = np.random.rand(n_channels, n_samples).astype(np.float32)
    snapshot = common_pb2.HDSnapshot()
    snapshot.priSignals.CopyFrom(
        common_pb2.RawSignals(
            data=data.tobytes(),
            shape=[n_channels, n_samples],
        )
    )
    return snapshot


def test_get_signals_lookback() -> None:
    """get_signals with FROM_HEAD mode sends correct request and returns SignalArrays."""
    with (
        patch("radiens_core._base_client.AuthInterceptor"),
        patch("radiens_core.allego_client.get_signal_group_proto"),
        patch(
            "radiens_core.allego_client.channel_metadata_from_proto",
            return_value=_make_channel_metadata(),
        ),
        patch(
            "radiens_core.allego_client.get_hd_snapshot_proto",
            return_value=_make_hd_snapshot(1, 30000),
        ) as mock_rpc,
    ):
        client = AllegoClient()
        sigs = client.get_signals(
            time_range=TimeRangeSpec.lookback(1.0),
            signals=[0],
        )

    assert isinstance(sigs, SignalArrays)
    assert sigs.amp is not None
    assert sigs.amp.shape == (1, 30000)

    mock_rpc.assert_called_once()
    args, _ = mock_rpc.call_args
    req: common_pb2.HDSnapshotRequest = args[2]
    assert req.streamGroupId == PRIMARY_CACHE_STREAM_GROUP_ID
    assert list(req.timeRange) == [1.0, float("nan")] or (req.timeRange[0] == 1.0 and np.isnan(req.timeRange[1]))
    assert list(req.priSignals.ntvChanIdxs) == [0]


def test_get_signals_subset() -> None:
    """get_signals with SUBSET mode sends [start, end] timeRange."""
    with (
        patch("radiens_core._base_client.AuthInterceptor"),
        patch("radiens_core.allego_client.get_signal_group_proto"),
        patch(
            "radiens_core.allego_client.channel_metadata_from_proto",
            return_value=_make_channel_metadata(),
        ),
        patch(
            "radiens_core.allego_client.get_hd_snapshot_proto",
            return_value=_make_hd_snapshot(1, 30000),
        ) as mock_rpc,
    ):
        client = AllegoClient()
        sigs = client.get_signals(time_range=[0.0, 1.0], signals=[0])

    assert isinstance(sigs, SignalArrays)
    args, _ = mock_rpc.call_args
    req: common_pb2.HDSnapshotRequest = args[2]
    assert list(req.timeRange) == [0.0, 1.0]
    assert req.streamGroupId == PRIMARY_CACHE_STREAM_GROUP_ID


def test_get_signals_lookback_shorthand() -> None:
    """get_signals accepts [duration, nan] shorthand and sends FROM_HEAD-style request."""
    with (
        patch("radiens_core._base_client.AuthInterceptor"),
        patch("radiens_core.allego_client.get_signal_group_proto"),
        patch(
            "radiens_core.allego_client.channel_metadata_from_proto",
            return_value=_make_channel_metadata(),
        ),
        patch(
            "radiens_core.allego_client.get_hd_snapshot_proto",
            return_value=_make_hd_snapshot(1, 30000),
        ) as mock_rpc,
    ):
        client = AllegoClient()
        sigs = client.get_signals(time_range=[1.0, float("nan")], signals=[0])

    assert isinstance(sigs, SignalArrays)
    args, _ = mock_rpc.call_args
    req: common_pb2.HDSnapshotRequest = args[2]
    assert req.timeRange[0] == 1.0
    assert np.isnan(req.timeRange[1])



"""CurateClient — data curation client."""

from __future__ import annotations

import os
import uuid
from typing import TYPE_CHECKING

from tqdm import tqdm

from radiens_core._base_client import BaseClient
from radiens_core._converters._dataset import file_type_to_proto, guess_file_type
from radiens_core._generated import common_pb2
from radiens_core._services._radiens_core import apply_protocol_proto, set_datasource, set_protocol_proto
from radiens_core._services._radiens_lifecycle import healthcheck_proto
from radiens_core._transport._channel_pool import ChannelPool
from radiens_core._transport._discovery import discover_radiens_addresses
from radiens_core.exceptions import RadiensError
from radiens_core.models.common import ClientType, ServiceEndpoint
from radiens_core.models.dataset import DatasetMetadata

if TYPE_CHECKING:
    from pathlib import Path

    from radiens_core.models.types import DatasetRef


class CurateClient(BaseClient):
    """Client for data curation operations.

    Auto-discovers a running ``radiensserver`` on the local machine.

    Usage::

        with CurateClient() as client:
            meta = client.link_data_file("/path/to/recording.xdat")
            new_meta = client.slice_time(meta, 0, 10, "sliced.xdat")
    """

    def __init__(self) -> None:
        tmp_pool = ChannelPool()
        try:
            addresses = discover_radiens_addresses(tmp_pool)
        finally:
            tmp_pool.close()

        core_addr = addresses.get(ServiceEndpoint.RADIENS_CORE, "")
        super().__init__(
            addresses=addresses,
            auth_server_address=core_addr,
            fallback_service=ServiceEndpoint.RADIENS_CORE,
        )
        self._datasets: dict[str, DatasetMetadata] = {}

    # --- Internal helpers ---

    def _resolve_dataset(self, ref: DatasetRef) -> DatasetMetadata:
        """Resolve a dataset reference to DatasetMetadata."""
        if isinstance(ref, DatasetMetadata):
            return ref
        ref_str = str(ref)
        if ref_str in self._datasets:
            return self._datasets[ref_str]
        raise ValueError(f"Unknown dataset: {ref_str}. Link it first with link_data_file().")

    def _build_linear_protocol(
        self,
        input_dataset: DatasetMetadata,
        transform_node: common_pb2.TransformNode,
        output_path: str | Path,
    ) -> tuple[list[common_pb2.TransformNode], list[common_pb2.TransformEdge]]:
        """Build a standard 3-node linear protocol: Source -> Transform -> Sink."""
        output_path_str = str(output_path)
        base_name = os.path.splitext(os.path.basename(output_path_str))[0]
        dir_path = os.path.dirname(output_path_str) or "."

        source_id = str(uuid.uuid4())
        sink_id = str(uuid.uuid4())

        # NOTE: 'invalid' is a 'required' field in the proto definition.
        # We must set it to False to pass client-side serialization.
        source_node = common_pb2.TransformNode(
            id=source_id,
            type=common_pb2.SOURCE,
            invalid=False,
            datasourceSinkParams=common_pb2.DataSourceSinkTransformParams(
                id=input_dataset.id,
                dsrcName=input_dataset.base_name,
                path=input_dataset.path,
                fileType=file_type_to_proto(input_dataset.file_type),
            ),
        )

        sink_node = common_pb2.TransformNode(
            id=sink_id,
            type=common_pb2.SINK,
            invalid=False,
            datasourceSinkParams=common_pb2.DataSourceSinkTransformParams(
                dsrcName=base_name,
                path=dir_path,
                fileType=file_type_to_proto(guess_file_type(output_path_str)),
            ),
        )

        transform_node.invalid = False

        edges = [
            common_pb2.TransformEdge(id=str(uuid.uuid4()), source=source_id, target=transform_node.id),
            common_pb2.TransformEdge(id=str(uuid.uuid4()), source=transform_node.id, target=sink_id),
        ]

        return [source_node, transform_node, sink_node], edges

    def _execute_protocol(
        self,
        nodes: list[common_pb2.TransformNode],
        edges: list[common_pb2.TransformEdge],
        desc: str = "Processing",
    ) -> None:
        """Submit and execute a protocol with a progress bar."""
        protocol_id = str(uuid.uuid4())
        protocol = common_pb2.Protocol(
            id=protocol_id,
            transformNodes=nodes,
            transformEdges=edges,
        )

        addr = self._server_address(ServiceEndpoint.RADIENS_CORE)
        validated_protocol = set_protocol_proto(self._pool, addr, protocol)

        # Check for validation errors from the server
        for node in validated_protocol.transformNodes:
            if node.invalid:
                raise RadiensError(f"Curation task failed validation at node '{node.id}': {node.errorMessage}")

        req = common_pb2.ProtocolRequest(protocolID=protocol_id)
        stream = apply_protocol_proto(self._pool, addr, req)

        with tqdm(total=100, desc=desc, unit="%") as pbar:
            last_frac = 0.0
            for progress in stream:
                # progress.fracComplete is 0.0 to 1.0
                inc = (progress.fracComplete - last_frac) * 100
                if inc > 0:
                    pbar.update(inc)
                    last_frac = progress.fracComplete
            # Ensure it reaches 100% if the stream finished successfully
            if last_frac < 1.0:
                pbar.update((1.0 - last_frac) * 100)

    # --- Public API ---

    def healthcheck(self, service: ServiceEndpoint = ServiceEndpoint.RADIENS_LIFECYCLE) -> None:
        """Verify the radiensserver is responsive.

        Parameters
        ----------
        service
            Which gRPC service to probe (default ``RADIENS_LIFECYCLE``).
            ``RADIENS_CORE`` is also supported.

        Raises
        ------
        RadiensError
            If the health check fails or the server is unresponsive.
        """
        addr = self._server_address(service)
        response = healthcheck_proto(self._pool, addr, ClientType.CURATE, service=service)
        if response.errMsg:
            raise RadiensError(f"Health check failed: {response.errMsg}")

    def link_data_file(
        self,
        path: str | Path,
    ) -> DatasetMetadata:
        """Link a recording file and return its metadata.

        Parameters
        ----------
        path
            Path to the recording file (e.g., .xdat, .rhd).

        Returns
        -------
        DatasetMetadata
            Metadata for the linked dataset.
        """
        addr = self._server_address(ServiceEndpoint.RADIENS_CORE)
        meta = set_datasource(self._pool, addr, str(path))
        self._datasets[meta.id] = meta
        return meta

    def slice_time(
        self,
        dataset: DatasetRef,
        start_sec: float,
        end_sec: float,
        output_path: str | Path,
    ) -> DatasetMetadata:
        """Slice a dataset in time.

        Parameters
        ----------
        dataset
            Dataset to slice (metadata or ID).
        start_sec
            Start time in seconds.
        end_sec
            End time in seconds.
        output_path
            Where to save the result.

        Returns
        -------
        DatasetMetadata
            Metadata for the new (sliced) dataset.
        """
        meta = self._resolve_dataset(dataset)

        transform_id = str(uuid.uuid4())
        transform_node = common_pb2.TransformNode(
            id=transform_id,
            type=common_pb2.SLICE_TIME,
            invalid=False,
            sliceTimeParams=common_pb2.SliceTimeTransformParams(
                timeStart=start_sec,
                timeEnd=end_sec,
            ),
        )

        nodes, edges = self._build_linear_protocol(meta, transform_node, output_path)
        self._execute_protocol(nodes, edges, desc=f"Slicing {os.path.basename(str(output_path))}")

        return self.link_data_file(output_path)

    def downsample(
        self,
        dataset: DatasetRef,
        factor: int,
        output_path: str | Path,
    ) -> DatasetMetadata:
        """Downsample a dataset.

        Parameters
        ----------
        dataset
            Dataset to downsample (metadata or ID).
        factor
            Downsampling factor (e.g., 2, 4, 10).
        output_path
            Where to save the result.

        Returns
        -------
        DatasetMetadata
            Metadata for the new (downsampled) dataset.
        """
        meta = self._resolve_dataset(dataset)

        transform_id = str(uuid.uuid4())
        transform_node = common_pb2.TransformNode(
            id=transform_id,
            type=common_pb2.DOWNSAMPLE,
            invalid=False,
            downsampleParams=common_pb2.DownsampleTransformParams(
                sampleFactor=factor,
            ),
        )

        nodes, edges = self._build_linear_protocol(meta, transform_node, output_path)
        self._execute_protocol(nodes, edges, desc=f"Downsampling {os.path.basename(str(output_path))}")

        return self.link_data_file(output_path)



"""Type aliases for the public API.

Uses Python 3.12 ``type`` statement. These will be replaced with
precise union types as the model modules are built out.
"""

from pathlib import Path

import numpy as np
import numpy.typing as npt

from radiens_core.models.dataset import DatasetMetadata
from radiens_core.models.signals import KeyIndex

# These are defined inline to avoid forward-reference import issues.
# They will be progressively tightened as model modules land.
type DatasetRef = str | Path | DatasetMetadata
"""Reference to a dataset: file path string or Path (DatasetMetadata added later)."""

type TimeRangeInput = list[float] | npt.NDArray[np.float64] | float | None
"""Any accepted time range format (TimeRangeSpec, TimeRange added later)."""

type ChannelIndices = list[int] | npt.NDArray[np.int64] | None
"""Channel index specification."""

type KeyIdxArrays = dict[KeyIndex, npt.NDArray[np.int64]]
"""Mapping from KeyIndex to channel index arrays."""



"""RadiensCore RPC wrappers — thin service layer (proto only)."""

from __future__ import annotations

from typing import TYPE_CHECKING

import grpc

from radiens_core._converters._dataset import file_type_to_proto, guess_file_type
from radiens_core._generated import (
    common_pb2,
    datasource_pb2,
    radiensserver_pb2_grpc,
)
from radiens_core._transport._errors import handle_grpc_error
from radiens_core.models.common import ClientType
from radiens_core.models.dataset import RadiensFileType

if TYPE_CHECKING:
    from collections.abc import Iterable

    from radiens_core._transport._channel_pool import ChannelPool
    from radiens_core.models.dataset import DatasetMetadata


def get_hd_snapshot_py_proto(
    pool: ChannelPool,
    addr: str,
    request: common_pb2.HDSnapshotRequest2,
) -> common_pb2.HDSnapshot2:
    """Send ``GetHDSnapshotPy`` RPC, return raw proto response."""
    channel = pool.get_channel(addr)
    stub = radiensserver_pb2_grpc.RadiensCoreStub(channel)
    try:
        return stub.GetHDSnapshotPy(request)
    except grpc.RpcError as ex:
        handle_grpc_error(ex, ClientType.VIDERE)
        raise  # unreachable, but satisfies mypy


def set_datasource(
    pool: ChannelPool,
    addr: str,
    path: str,
) -> DatasetMetadata:
    """Link a data file via SetDataSourceFromFile RPC.

    Returns domain DatasetMetadata.
    """
    import os

    from radiens_core._converters._dataset import dataset_metadata_from_proto

    ft = guess_file_type(path)
    base_name = os.path.splitext(os.path.basename(path))[0]

    # Normalize XDAT base name by stripping standard suffixes if present
    if ft == RadiensFileType.XDAT:
        if base_name.endswith("_data"):
            base_name = base_name[:-5]
        elif base_name.endswith("_timestamp"):
            base_name = base_name[:-10]

    dir_path = os.path.dirname(path) or "."

    req = datasource_pb2.DataSourceSetSaveRequest(
        path=dir_path,
        baseName=base_name,
        fileType=file_type_to_proto(ft),
    )

    channel = pool.get_channel(addr)
    stub = radiensserver_pb2_grpc.RadiensCoreStub(channel)
    try:
        res = stub.SetDataSourceFromFile(req)
    except grpc.RpcError as ex:
        handle_grpc_error(ex, ClientType.VIDERE)
        raise  # unreachable, but satisfies mypy

    return dataset_metadata_from_proto(res)


def set_protocol_proto(
    pool: ChannelPool,
    addr: str,
    request: common_pb2.Protocol,
) -> common_pb2.Protocol:
    """Send ``SetProtocol`` RPC, return raw proto response."""
    channel = pool.get_channel(addr)
    stub = radiensserver_pb2_grpc.RadiensCoreStub(channel)
    try:
        return stub.SetProtocol(request)
    except grpc.RpcError as ex:
        handle_grpc_error(ex, ClientType.CURATE)
        raise  # unreachable, but satisfies mypy


def apply_protocol_proto(
    pool: ChannelPool,
    addr: str,
    request: common_pb2.ProtocolRequest,
) -> Iterable[common_pb2.ApplyProtocolProgress]:
    """Send ``ApplyProtocol`` RPC, return stream of ApplyProtocolProgress."""
    channel = pool.get_channel(addr)
    stub = radiensserver_pb2_grpc.RadiensCoreStub(channel)
    try:
        return stub.ApplyProtocol(request)
    except grpc.RpcError as ex:
        handle_grpc_error(ex, ClientType.CURATE)
        raise  # unreachable, but satisfies mypy



"""DatasetMetadata — immutable metadata for a linked dataset."""

from __future__ import annotations

from collections.abc import Sequence
from enum import IntEnum
from typing import Annotated

import numpy as np
import numpy.typing as npt
from pydantic import BaseModel, BeforeValidator, ConfigDict

from radiens_core.models.channels import ChannelMetadata
from radiens_core.models.time_range import TimeRange


class TrsMode(IntEnum):
    """Time range selection mode. Values match proto ``common.TimeRangeSelMode``."""

    SUBSET = 0  # TRS_SUBSET
    TO_HEAD = 1  # TRS_TO_HEAD
    FROM_HEAD = 2  # TRS_FROM_HEAD


class RadiensFileType(IntEnum):
    """Recording file type. Values match proto ``common.RadixFileTypes``."""

    RHD = 0
    XDAT = 1
    CSV = 2
    HDF5 = 3
    NEX5 = 4
    NWB = 5
    KILOSORT2 = 8
    NSX = 13
    TDT = 15
    SPIKES = 16
    OE_DAT = 22


class DatasetMetadata(BaseModel):
    """Metadata for a linked dataset.

    Returned by ``link_data_file()`` and similar operations.
    Contains everything needed to query signals from this dataset.
    """

    model_config = ConfigDict(frozen=True)

    id: str
    base_name: str
    path: str
    file_type: RadiensFileType
    time_range: TimeRange
    channel_metadata: ChannelMetadata
    parent_dsource_id: str = ""
    probe_uid: str = ""

    @property
    def duration(self) -> float:
        """Dataset duration in seconds."""
        return self.time_range.dur_sec

    @property
    def sampling_rate(self) -> float:
        """Sampling rate in Hz."""
        return self.time_range.fs


class TimeRangeSpec(BaseModel):
    """Semantic time range specification with embedded mode.

    Use factory methods to create specs with clear semantics:

    >>> spec = TimeRangeSpec.subset(0, 10)     # [0, 10) seconds
    >>> spec = TimeRangeSpec.lookback(5.0)      # Last 5 seconds
    >>> spec = TimeRangeSpec.full()              # Entire recording
    """

    model_config = ConfigDict(frozen=True)

    mode: TrsMode
    start: float
    end: float | None = None

    # --- Factory methods ---

    @classmethod
    def subset(cls, start: float, end: float) -> TimeRangeSpec:
        """Absolute time range ``[start, end)`` seconds."""
        if end <= start:
            raise ValueError(f"end ({end}) must be greater than start ({start})")
        return cls(mode=TrsMode.SUBSET, start=start, end=end)

    @classmethod
    def lookback(cls, duration: float) -> TimeRangeSpec:
        """Last ``duration`` seconds from head of cache."""
        if duration <= 0:
            raise ValueError(f"duration must be positive, got {duration}")
        return cls(mode=TrsMode.FROM_HEAD, start=duration, end=None)

    @classmethod
    def to_head(cls, start: float) -> TimeRangeSpec:
        """From ``start`` seconds to current head."""
        return cls(mode=TrsMode.TO_HEAD, start=start, end=None)

    @classmethod
    def full(cls) -> TimeRangeSpec:
        """Full dataset/cache range."""
        return cls(mode=TrsMode.SUBSET, start=0, end=float("inf"))

    @classmethod
    def from_list(cls, time_range: list[float], mode: TrsMode | None = None) -> TimeRangeSpec:
        """Create from ``[start, end]`` or ``[duration, nan]`` list."""
        arr = np.asarray(time_range, dtype=np.float64)
        if len(arr) != 2:
            raise ValueError("time_range must have exactly 2 elements")
        if mode is None:
            mode = TrsMode.FROM_HEAD if np.isnan(arr[1]) else TrsMode.SUBSET
        if mode == TrsMode.FROM_HEAD:
            return cls(mode=mode, start=float(arr[0]), end=None)
        return cls(mode=mode, start=float(arr[0]), end=float(arr[1]))

    # --- Resolution methods ---

    def is_full(self) -> bool:
        """Check if this is a full-range specification."""
        return self.mode == TrsMode.SUBSET and self.end == float("inf")

    def to_array(self) -> npt.NDArray[np.float64]:
        """Convert to ``[start, end]`` or ``[duration, nan]`` array."""
        if self.is_full():
            raise ValueError("Cannot convert full() spec to array without dataset.")
        if self.mode in (TrsMode.FROM_HEAD, TrsMode.TO_HEAD):
            return np.array([self.start, np.nan], dtype=np.float64)
        return np.array([self.start, self.end], dtype=np.float64)

    def resolve(self, dataset_metadata: object | None = None, fs: float | None = None) -> TimeRange | None:
        """Resolve to a concrete TimeRange.

        Parameters
        ----------
        dataset_metadata
            Dataset metadata used to obtain ``fs`` and dataset bounds.
        fs
            Sampling rate override (used when ``dataset_metadata`` is not provided).

        Returns
        -------
        TimeRange or None
            ``None`` for ``FROM_HEAD`` mode — caller must use ``to_array()`` instead.
        """
        if self.mode == TrsMode.FROM_HEAD:
            return None

        if not isinstance(dataset_metadata, DatasetMetadata):
            raise ValueError("dataset_metadata required to resolve time range")

        dataset_tr = dataset_metadata.time_range
        effective_fs = fs if fs is not None else dataset_tr.fs

        if self.is_full():
            return dataset_tr

        if self.mode == TrsMode.SUBSET:
            end = self.end if (self.end is not None and self.end != float("inf")) else dataset_tr.sec[1]
            ts_start = int(self.start * effective_fs)
            ts_end = int(end * effective_fs)
            return TimeRange(
                sec=(self.start, end),
                timestamp=(ts_start, ts_end),
                fs=effective_fs,
                dur_sec=end - self.start,
                walltime_str=dataset_tr.walltime_str,
                n=ts_end - ts_start,
            )

        if self.mode == TrsMode.TO_HEAD:
            ts_start = int(self.start * effective_fs)
            ts_end = dataset_tr.timestamp[1]
            end = dataset_tr.sec[1]
            return TimeRange(
                sec=(self.start, end),
                timestamp=(ts_start, ts_end),
                fs=effective_fs,
                dur_sec=end - self.start,
                walltime_str=dataset_tr.walltime_str,
                n=ts_end - ts_start,
            )

        raise ValueError(f"Cannot resolve time range with mode: {self.mode}")

    @property
    def duration(self) -> float | None:
        """Duration in seconds, if known."""
        if self.mode == TrsMode.FROM_HEAD:
            return self.start
        if self.mode == TrsMode.SUBSET and self.end is not None and self.end != float("inf"):
            return self.end - self.start
        return None

    def __repr__(self) -> str:
        if self.is_full():
            return "TimeRangeSpec.full()"
        if self.mode == TrsMode.FROM_HEAD:
            return f"TimeRangeSpec.lookback({self.start})"
        if self.mode == TrsMode.TO_HEAD:
            return f"TimeRangeSpec.to_head({self.start})"
        return f"TimeRangeSpec.subset({self.start}, {self.end})"


def _coerce_time_range(v: object) -> TimeRangeSpec:
    """Coerce raw Python types into :class:`TimeRangeSpec`.

    Accepted inputs:
    - ``TimeRangeSpec`` — pass-through
    - ``list[float]`` / ``tuple[float, ...]`` — delegates to ``TimeRangeSpec.from_list``
    """
    if isinstance(v, TimeRangeSpec):
        return v
    if isinstance(v, (list, tuple)):
        return TimeRangeSpec.from_list(list(v))
    raise ValueError(f"Expected TimeRangeSpec, list, or tuple — got {type(v).__name__}")


FlexTimeRange = Annotated[
    TimeRangeSpec | Sequence[float],
    BeforeValidator(_coerce_time_range),
]
"""``TimeRangeSpec`` that also accepts ``list``/``tuple`` shorthand.

Use in Pydantic models::

    class MyRequest(BaseModel):
        time_range: FlexTimeRange
"""



"""Tests for CurateClient."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from collections.abc import Generator
from unittest.mock import MagicMock, patch

import pytest

from radiens_core._generated import common_pb2
from radiens_core.curate_client import CurateClient
from radiens_core.models.dataset import RadiensFileType


@pytest.fixture
def mock_discovery() -> Generator[Any, None, None]:  # type: ignore[explicit-any]
    """Mock server discovery."""
    with patch("radiens_core.curate_client.discover_radiens_addresses") as mock:
        mock.return_value = {
            "RADIENS_CORE": "localhost:50055",
            "RADIENS_LIFECYCLE": "localhost:50055",
        }
        yield mock


@pytest.fixture
def mock_auth() -> Generator[Any, None, None]:  # type: ignore[explicit-any]
    """Mock authentication interceptor."""
    with patch("radiens_core._base_client.AuthInterceptor") as mock:
        yield mock


def test_execute_protocol_progress(mock_discovery: Any, mock_auth: Any) -> None:  # type: ignore[explicit-any]
    """Test _execute_protocol with a mocked progress stream."""
    client = CurateClient()

    # Mock set_protocol_proto
    mock_set = MagicMock()

    # Mock apply_protocol_proto to return a stream of progress messages
    mock_apply = MagicMock()

    def progress_stream(*args: Any, **kwargs: Any) -> Generator[common_pb2.ApplyProtocolProgress, None, None]:  # type: ignore[explicit-any]
        yield common_pb2.ApplyProtocolProgress(fracComplete=0.2)
        yield common_pb2.ApplyProtocolProgress(fracComplete=0.5)
        yield common_pb2.ApplyProtocolProgress(fracComplete=1.0)

    mock_apply.return_value = progress_stream()

    with (
        patch("radiens_core.curate_client.set_protocol_proto", mock_set),
        patch("radiens_core.curate_client.apply_protocol_proto", mock_apply),
        patch("radiens_core.curate_client.tqdm") as mock_tqdm,
    ):
        # Mock tqdm instance
        pbar = mock_tqdm.return_value.__enter__.return_value

        nodes = [common_pb2.TransformNode(id="node1")]
        edges = [common_pb2.TransformEdge(id="edge1")]

        client._execute_protocol(nodes, edges, desc="Test Task")

        # Verify set_protocol was called
        mock_set.assert_called_once()
        protocol = mock_set.call_args[0][2]
        assert protocol.transformNodes[0].id == "node1"
        assert protocol.transformEdges[0].id == "edge1"

        # Verify apply_protocol was called
        mock_apply.assert_called_once()

        # Verify tqdm updates
        # Expected updates: 0.2*100=20, (0.5-0.2)*100=30, (1.0-0.5)*100=50
        assert pbar.update.call_count == 3
        calls = pbar.update.call_args_list
        assert calls[0][0][0] == pytest.approx(20.0)
        assert calls[1][0][0] == pytest.approx(30.0)
        assert calls[2][0][0] == pytest.approx(50.0)


def test_link_data_file(mock_discovery: Any, mock_auth: Any) -> None:  # type: ignore[explicit-any]
    """Test link_data_file method calls the correct service."""
    client = CurateClient()
    mock_meta = MagicMock()
    mock_meta.id = "test_ds"

    with patch("radiens_core.curate_client.set_datasource", return_value=mock_meta) as mock_set_ds:
        meta = client.link_data_file("/path/to/test.xdat")

        assert meta == mock_meta
        assert client._datasets["test_ds"] == mock_meta
        mock_set_ds.assert_called_once()
        assert mock_set_ds.call_args[0][2] == "/path/to/test.xdat"


def test_slice_time(mock_discovery: Any, mock_auth: Any) -> None:  # type: ignore[explicit-any]
    """Test slice_time method builds correct graph and executes."""
    client = CurateClient()

    mock_input_meta = MagicMock()
    mock_input_meta.id = "input_ds"
    mock_input_meta.base_name = "test_recording"
    mock_input_meta.path = "/path/to"
    mock_input_meta.file_type = RadiensFileType.XDAT
    client._datasets["input_ds"] = mock_input_meta

    mock_output_meta = MagicMock()

    with (
        patch.object(client, "_execute_protocol") as mock_exec,
        patch.object(client, "link_data_file", return_value=mock_output_meta),
    ):
        res = client.slice_time("input_ds", 10.5, 20.5, "output.xdat")

        assert res == mock_output_meta
        mock_exec.assert_called_once()
        args, kwargs = mock_exec.call_args
        nodes, edges = args
        assert "Slicing output.xdat" in kwargs["desc"]

        # Verify nodes: Source, Transform, Sink
        assert len(nodes) == 3
        assert nodes[0].type == common_pb2.SOURCE
        assert nodes[0].datasourceSinkParams.id == "input_ds"

        assert nodes[1].type == common_pb2.SLICE_TIME
        assert nodes[1].sliceTimeParams.timeStart == 10.5
        assert nodes[1].sliceTimeParams.timeEnd == 20.5

        assert nodes[2].type == common_pb2.SINK
        assert nodes[2].datasourceSinkParams.dsrcName == "output"

        # Verify edges
        assert len(edges) == 2
        assert edges[0].source == nodes[0].id
        assert edges[0].target == nodes[1].id
        assert edges[1].source == nodes[1].id
        assert edges[1].target == nodes[2].id


def test_downsample(mock_discovery: Any, mock_auth: Any) -> None:  # type: ignore[explicit-any]
    """Test downsample method builds correct graph and executes."""
    client = CurateClient()

    mock_input_meta = MagicMock()
    mock_input_meta.id = "input_ds"
    mock_input_meta.base_name = "test_recording"
    mock_input_meta.path = "/path/to"
    mock_input_meta.file_type = RadiensFileType.XDAT
    client._datasets["input_ds"] = mock_input_meta

    mock_output_meta = MagicMock()

    with (
        patch.object(client, "_execute_protocol") as mock_exec,
        patch.object(client, "link_data_file", return_value=mock_output_meta),
    ):
        res = client.downsample("input_ds", 4, "downsampled.rhd")

        assert res == mock_output_meta
        mock_exec.assert_called_once()
        args, kwargs = mock_exec.call_args
        nodes, _edges = args
        assert "Downsampling downsampled.rhd" in kwargs["desc"]

        assert len(nodes) == 3
        assert nodes[1].type == common_pb2.DOWNSAMPLE
        assert nodes[1].downsampleParams.sampleFactor == 4

        assert nodes[2].datasourceSinkParams.fileType == common_pb2.RHD



"""Domain models — pure Pydantic v2, zero proto dependency.

All enums and models are importable from their domain-grouped submodules:

    from radiens_core.models.stim import StimParams, StimShape
    from radiens_core.models.signals import SignalType, SigSelect
    from radiens_core.models.status import RecordMode, StreamMode
"""

from radiens_core.models.channels import (
    AreaUnits,
    ChannelInfo,
    ChannelMetadata,
    KeyIdxs,
    SignalUnits,
    SitePosition,
)
from radiens_core.models.common import ClientType, ServiceEndpoint
from radiens_core.models.dataset import DatasetMetadata, RadiensFileType, TrsMode
from radiens_core.models.ports import GPIOChannelCount, Port, PortChannelCount
from radiens_core.models.signals import KeyIndex, SignalArrays, SignalType, SigSelect
from radiens_core.models.status import (
    BackboneMode,
    BackboneStatus,
    DataRecordingType,
    RecordingConfig,
    RecordingStatus,
    RecordMode,
    RecordStateRequest,
    RestartRequest,
    StreamingStatus,
    StreamMode,
    StreamStateRequest,
)
from radiens_core.models.stim import (
    StimKeypressIndex,
    StimParams,
    StimPolarity,
    StimPulseMode,
    StimShape,
    StimStep,
    StimStepRequest,
    StimStepState,
    StimTriggerEdgeOrLevel,
    StimTriggerHighOrLow,
)
from radiens_core.models.time_range import TimeRange
from radiens_core.models.triggers import (
    ManualStimTriggerRequest,
    ManualStimTriggerToggleRequest,
    SetTriggerStateRequest,
    TriggerChannel,
    TriggerState,
)

__all__ = [
    "AreaUnits",
    "BackboneMode",
    "BackboneStatus",
    "ChannelInfo",
    "ChannelMetadata",
    "ClientType",
    "DataRecordingType",
    "DatasetMetadata",
    "GPIOChannelCount",
    "KeyIdxs",
    "KeyIndex",
    "ManualStimTriggerRequest",
    "ManualStimTriggerToggleRequest",
    "Port",
    "PortChannelCount",
    "RadiensFileType",
    "RecordMode",
    "RecordStateRequest",
    "RecordingConfig",
    "RecordingStatus",
    "RestartRequest",
    "ServiceEndpoint",
    "SetTriggerStateRequest",
    "SigSelect",
    "SignalArrays",
    "SignalType",
    "SignalUnits",
    "SitePosition",
    "StimKeypressIndex",
    "StimParams",
    "StimPolarity",
    "StimPulseMode",
    "StimShape",
    "StimStep",
    "StimStepRequest",
    "StimStepState",
    "StimTriggerEdgeOrLevel",
    "StimTriggerHighOrLow",
    "StreamMode",
    "StreamStateRequest",
    "StreamingStatus",
    "TimeRange",
    "TriggerChannel",
    "TriggerState",
    "TrsMode",
]



"""Restart converters — domain models <-> proto messages."""

from __future__ import annotations

from typing import TYPE_CHECKING

from radiens_core._generated import allegoserver_pb2
from radiens_core.models.status import BackboneMode

if TYPE_CHECKING:
    from radiens_core.models.status import RestartRequest


# --- BackboneMode enum conversion ---

_BACKBONE_MODE_TO_PROTO: dict[BackboneMode, allegoserver_pb2.BackboneMode.ValueType] = {
    BackboneMode.SMARTBOX_PRO: allegoserver_pb2.SMARTBOX_PRO,
    BackboneMode.SMARTBOX_SIM_GEN_SINE: allegoserver_pb2.SMARTBOX_SIM_GEN_SINE,
    BackboneMode.SMARTBOX_SIM_GEN_SPIKES: allegoserver_pb2.SMARTBOX_SIM_GEN_SPIKES,
    BackboneMode.SMARTBOX_SIM_DATASOURCE: allegoserver_pb2.SMARTBOX_SIM_DATASOURCE,
    BackboneMode.OPEN_EPHYS_USB3: allegoserver_pb2.OPEN_EPHYS_USB3,
    BackboneMode.INTAN_RECORDING_CONTROLLER_1024: allegoserver_pb2.INTAN_RECORDING_CONTROLLER_1024,
    BackboneMode.SMARTBOX_CLASSIC: allegoserver_pb2.SMARTBOX_CLASSIC,
    BackboneMode.INTAN_RECORDING_CONTROLLER_512: allegoserver_pb2.INTAN_RECORDING_CONTROLLER_512,
    BackboneMode.OPEN_EPHYS_USB2: allegoserver_pb2.OPEN_EPHYS_USB2,
    BackboneMode.INTAN_USB2: allegoserver_pb2.INTAN_USB2,
    BackboneMode.SMARTBOX_SIM_GEN_SINE_MAPPED: allegoserver_pb2.SMARTBOX_SIM_GEN_SINE_MAPPED,
    BackboneMode.SMARTBOX_SIM_GEN_SINE_HIGH_FREQ: allegoserver_pb2.SMARTBOX_SIM_GEN_SINE_HIGH_FREQ,
    BackboneMode.SMARTBOX_SIM_GEN_SINE_MULTI_BAND: allegoserver_pb2.SMARTBOX_SIM_GEN_SINE_MULTI_BAND,
    BackboneMode.SMARTBOX_PRO_SINAPS_256: allegoserver_pb2.SMARTBOX_PRO_SINAPS_256,
    BackboneMode.SMARTBOX_PRO_SINAPS_1024: allegoserver_pb2.SMARTBOX_PRO_SINAPS_1024,
    BackboneMode.XDAQ_ONE_REC: allegoserver_pb2.XDAQ_ONE_REC,
    BackboneMode.XDAQ_ONE_STIM: allegoserver_pb2.XDAQ_ONE_STIM,
    BackboneMode.XDAQ_CORE_REC: allegoserver_pb2.XDAQ_CORE_REC,
    BackboneMode.XDAQ_CORE_STIM: allegoserver_pb2.XDAQ_CORE_STIM,
    BackboneMode.SMARTBOX_PRO_SINAPS_512: allegoserver_pb2.SMARTBOX_PRO_SINAPS_512,
    BackboneMode.XDAQ_GEN2_ONE_STIM: allegoserver_pb2.XDAQ_GEN2_ONE_STIM,
    BackboneMode.XDAQ_GEN2_CORE_STIM: allegoserver_pb2.XDAQ_GEN2_CORE_STIM,
    BackboneMode.XDAQ_GEN2_ONE_REC: allegoserver_pb2.XDAQ_GEN2_ONE_REC,
    BackboneMode.XDAQ_GEN2_CORE_REC: allegoserver_pb2.XDAQ_GEN2_CORE_REC,
    BackboneMode.SMARTBOX_SIM_GEN_LFP: allegoserver_pb2.SMARTBOX_SIM_GEN_LFP,
}


def backbone_mode_to_proto(bm: BackboneMode) -> allegoserver_pb2.BackboneMode.ValueType:
    """Convert domain BackboneMode to proto BackboneMode."""
    return _BACKBONE_MODE_TO_PROTO[bm]


_BACKBONE_MODE_FROM_PROTO: dict[int, BackboneMode] = {
    int(allegoserver_pb2.SMARTBOX_PRO): BackboneMode.SMARTBOX_PRO,
    int(allegoserver_pb2.SMARTBOX_SIM_GEN_SINE): BackboneMode.SMARTBOX_SIM_GEN_SINE,
    int(allegoserver_pb2.SMARTBOX_SIM_GEN_SPIKES): BackboneMode.SMARTBOX_SIM_GEN_SPIKES,
    int(allegoserver_pb2.SMARTBOX_SIM_DATASOURCE): BackboneMode.SMARTBOX_SIM_DATASOURCE,
    int(allegoserver_pb2.OPEN_EPHYS_USB3): BackboneMode.OPEN_EPHYS_USB3,
    int(allegoserver_pb2.INTAN_RECORDING_CONTROLLER_1024): BackboneMode.INTAN_RECORDING_CONTROLLER_1024,
    int(allegoserver_pb2.SMARTBOX_CLASSIC): BackboneMode.SMARTBOX_CLASSIC,
    int(allegoserver_pb2.INTAN_RECORDING_CONTROLLER_512): BackboneMode.INTAN_RECORDING_CONTROLLER_512,
    int(allegoserver_pb2.OPEN_EPHYS_USB2): BackboneMode.OPEN_EPHYS_USB2,
    int(allegoserver_pb2.INTAN_USB2): BackboneMode.INTAN_USB2,
    int(allegoserver_pb2.SMARTBOX_SIM_GEN_SINE_MAPPED): BackboneMode.SMARTBOX_SIM_GEN_SINE_MAPPED,
    int(allegoserver_pb2.SMARTBOX_SIM_GEN_SINE_HIGH_FREQ): BackboneMode.SMARTBOX_SIM_GEN_SINE_HIGH_FREQ,
    int(allegoserver_pb2.SMARTBOX_SIM_GEN_SINE_MULTI_BAND): BackboneMode.SMARTBOX_SIM_GEN_SINE_MULTI_BAND,
    int(allegoserver_pb2.SMARTBOX_PRO_SINAPS_256): BackboneMode.SMARTBOX_PRO_SINAPS_256,
    int(allegoserver_pb2.SMARTBOX_PRO_SINAPS_1024): BackboneMode.SMARTBOX_PRO_SINAPS_1024,
    int(allegoserver_pb2.XDAQ_ONE_REC): BackboneMode.XDAQ_ONE_REC,
    int(allegoserver_pb2.XDAQ_ONE_STIM): BackboneMode.XDAQ_ONE_STIM,
    int(allegoserver_pb2.XDAQ_CORE_REC): BackboneMode.XDAQ_CORE_REC,
    int(allegoserver_pb2.XDAQ_CORE_STIM): BackboneMode.XDAQ_CORE_STIM,
    int(allegoserver_pb2.SMARTBOX_PRO_SINAPS_512): BackboneMode.SMARTBOX_PRO_SINAPS_512,
    int(allegoserver_pb2.XDAQ_GEN2_ONE_STIM): BackboneMode.XDAQ_GEN2_ONE_STIM,
    int(allegoserver_pb2.XDAQ_GEN2_CORE_STIM): BackboneMode.XDAQ_GEN2_CORE_STIM,
    int(allegoserver_pb2.XDAQ_GEN2_ONE_REC): BackboneMode.XDAQ_GEN2_ONE_REC,
    int(allegoserver_pb2.XDAQ_GEN2_CORE_REC): BackboneMode.XDAQ_GEN2_CORE_REC,
    int(allegoserver_pb2.SMARTBOX_SIM_GEN_LFP): BackboneMode.SMARTBOX_SIM_GEN_LFP,
}


def backbone_mode_from_proto(pb: int) -> BackboneMode:
    """Convert proto BackboneMode int to domain BackboneMode."""
    return _BACKBONE_MODE_FROM_PROTO[pb]


def restart_request_to_proto(
    request: RestartRequest,
) -> allegoserver_pb2.RestartRequest:
    """Convert domain RestartRequest to proto RestartRequest."""
    return allegoserver_pb2.RestartRequest(
        mode=backbone_mode_to_proto(request.mode),
    )



"""Shared enum converters used across multiple converter modules.

Only enum converters that are imported by more than one converter file
belong here. Single-use converters should live in their domain converter.
"""

from __future__ import annotations

from radiens_core._generated import allegoserver_pb2, common_pb2
from radiens_core.models.ports import Port
from radiens_core.models.status import RecordMode, StreamMode

# --- Port ---

_PORT_TO_PROTO: dict[Port, common_pb2.Port.ValueType] = {
    Port.A: common_pb2.A,
    Port.B: common_pb2.B,
    Port.C: common_pb2.C,
    Port.D: common_pb2.D,
    Port.E: common_pb2.E,
    Port.F: common_pb2.F,
    Port.G: common_pb2.G,
    Port.H: common_pb2.H,
}

_PORT_FROM_PROTO: dict[int, Port] = {
    int(common_pb2.A): Port.A,
    int(common_pb2.B): Port.B,
    int(common_pb2.C): Port.C,
    int(common_pb2.D): Port.D,
    int(common_pb2.E): Port.E,
    int(common_pb2.F): Port.F,
    int(common_pb2.G): Port.G,
    int(common_pb2.H): Port.H,
}


def port_to_proto(port: Port) -> common_pb2.Port.ValueType:
    """Convert domain Port to proto Port."""
    return _PORT_TO_PROTO[port]


def port_from_proto(pb: int) -> Port:
    """Convert proto Port int to domain Port."""
    return _PORT_FROM_PROTO[pb]


# --- RecordMode ---

_RECORD_MODE_TO_PROTO: dict[RecordMode, allegoserver_pb2.RecordMode.ValueType] = {
    RecordMode.OFF: allegoserver_pb2.R_OFF,
    RecordMode.ON: allegoserver_pb2.R_ON,
}

_RECORD_MODE_FROM_PROTO: dict[int, RecordMode] = {
    int(allegoserver_pb2.R_OFF): RecordMode.OFF,
    int(allegoserver_pb2.R_ON): RecordMode.ON,
}


def record_mode_to_proto(rm: RecordMode) -> allegoserver_pb2.RecordMode.ValueType:
    """Convert domain RecordMode to proto RecordMode."""
    return _RECORD_MODE_TO_PROTO[rm]


def record_mode_from_proto(pb: int) -> RecordMode:
    """Convert proto RecordMode int to domain RecordMode."""
    return _RECORD_MODE_FROM_PROTO[pb]


# --- StreamMode ---

_STREAM_MODE_TO_PROTO: dict[StreamMode, allegoserver_pb2.StreamMode.ValueType] = {
    StreamMode.OFF: allegoserver_pb2.S_OFF,
    StreamMode.ON: allegoserver_pb2.S_ON,
}

_STREAM_MODE_FROM_PROTO: dict[int, StreamMode] = {
    int(allegoserver_pb2.S_OFF): StreamMode.OFF,
    int(allegoserver_pb2.S_ON): StreamMode.ON,
}


def stream_mode_to_proto(sm: StreamMode) -> allegoserver_pb2.StreamMode.ValueType:
    """Convert domain StreamMode to proto StreamMode."""
    return _STREAM_MODE_TO_PROTO[sm]


def stream_mode_from_proto(pb: int) -> StreamMode:
    """Convert proto StreamMode int to domain StreamMode."""
    return _STREAM_MODE_FROM_PROTO[pb]



"""AllegoClient — real-time data acquisition client."""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING, cast

from pydantic import ConfigDict, validate_call

from radiens_core._base_client import BaseClient
from radiens_core._constants import (
    ALLEGO_SERVICE_DEFAULTS,
    PRIMARY_CACHE_STREAM_GROUP_ID,
)
from radiens_core._converters._channels import channel_metadata_from_proto
from radiens_core._converters._record_state import record_state_request_to_proto
from radiens_core._converters._recording_config import recording_config_from_proto, recording_config_to_proto
from radiens_core._converters._restart import restart_request_to_proto
from radiens_core._converters._signals import signal_arrays_from_hd_snapshot
from radiens_core._converters._status import backbone_status_from_proto
from radiens_core._converters._stim import stim_params_from_proto, stim_params_to_proto
from radiens_core._converters._stim_step import stim_step_request_to_proto, stim_step_state_from_proto
from radiens_core._converters._stream_state import stream_state_request_to_proto
from radiens_core._converters._trigger_control import (
    manual_stim_trigger_request_to_proto,
    manual_stim_trigger_toggle_request_to_proto,
    set_trigger_state_request_to_proto,
    trigger_state_from_proto,
)
from radiens_core._generated import common_pb2
from radiens_core._services._allego_core import (
    get_recording_config_proto,
    get_signal_group_proto,
    get_status_proto,
    get_stim_params_proto,
    get_stim_step_proto,
    get_trigger_state_proto,
    healthcheck_proto,
    manual_stim_trigger_proto,
    manual_stim_trigger_toggle_proto,
    restart_proto,
    set_record_state_proto,
    set_recording_config_proto,
    set_stim_params_proto,
    set_stim_step_proto,
    set_stream_state_proto,
    set_trigger_state_proto,
)
from radiens_core._services._allego_pcache import get_hd_snapshot_proto
from radiens_core.exceptions import RadiensError
from radiens_core.models.common import ServiceEndpoint
from radiens_core.models.dataset import FlexTimeRange, TrsMode
from radiens_core.models.signals import FlexSignalSpec, SignalArrays

if TYPE_CHECKING:
    from radiens_core.models.channels import ChannelMetadata
    from radiens_core.models.dataset import TimeRangeSpec
    from radiens_core.models.signals import SignalSpec
    from radiens_core.models.status import (
        BackboneStatus,
        RecordingConfig,
        RecordStateRequest,
        RestartRequest,
        StreamStateRequest,
    )
    from radiens_core.models.stim import StimParams, StimStepRequest, StimStepState
    from radiens_core.models.triggers import (
        ManualStimTriggerRequest,
        ManualStimTriggerToggleRequest,
        SetTriggerStateRequest,
        TriggerState,
    )

_log = logging.getLogger(__name__)


def _time_range_floats(tr_spec: TimeRangeSpec) -> list[float]:
    """Convert a TimeRangeSpec to [start, end_or_nan] for HDSnapshotRequest.timeRange.

    FROM_HEAD: [duration_sec, nan] — server reads back ``duration`` seconds from cache head.
    SUBSET:    [start_sec, end_sec].
    TO_HEAD:   [start_sec, nan].
    """
    if tr_spec.mode == TrsMode.SUBSET:
        return [tr_spec.start, tr_spec.end if tr_spec.end is not None else 0.0]
    return [tr_spec.start, float("nan")]


class AllegoClient(BaseClient):
    """Client for real-time data acquisition and streaming.

    Auto-discovers running servers.  Falls back to well-known allego
    ports if server discovery is unavailable (no radiensserver running).

    Usage::

        with AllegoClient() as client:
            params = client.get_stim_params()
    """

    def __init__(self) -> None:
        addresses = self._discover()
        core_addr = addresses[ServiceEndpoint.ALLEGO_CORE]
        super().__init__(
            addresses=addresses,
            auth_server_address=core_addr,
            fallback_service=ServiceEndpoint.ALLEGO_CORE,
        )

    @staticmethod
    def _discover() -> dict[str, str]:
        return dict(ALLEGO_SERVICE_DEFAULTS)

    # --- Internal helpers ---

    @property
    def _core_address(self) -> str:
        return self._server_address(ServiceEndpoint.ALLEGO_CORE)

    @property
    def _pcache_address(self) -> str:
        return self._server_address(ServiceEndpoint.ALLEGO_PCACHE)

    # --- Public API ---

    def get_stim_params(self) -> dict[int, StimParams]:
        """Get stimulation parameters for all configured channels.

        Returns a dict keyed by system channel index (ntvAbsKeyIdx).
        """
        response = get_stim_params_proto(self._pool, self._core_address)
        return stim_params_from_proto(response)

    def get_status(self) -> BackboneStatus:
        """Get comprehensive system status.

        Returns current streaming status, recording status, port information,
        connection state, and GPIO channel counts.
        """
        response = get_status_proto(self._pool, self._core_address)
        return backbone_status_from_proto(response)

    def get_channel_metadata(self) -> ChannelMetadata:
        """Get current channel metadata for all connected channels.

        Returns:
            Complete channel metadata including probe geometry, signal types,
            channel indices, and site positions for all channels.

        Raises:
            RadiensError: On communication failure or server error
        """
        response = get_signal_group_proto(self._pool, self._core_address)
        return channel_metadata_from_proto(response)

    @validate_call(config=ConfigDict(arbitrary_types_allowed=True))
    def get_signals(
        self,
        time_range: FlexTimeRange,
        signals: FlexSignalSpec,
    ) -> SignalArrays:
        """Retrieve signal data from the primary live cache.

        Parameters
        ----------
        time_range
            Time range spec. Most common: ``TimeRangeSpec.lookback(5.0)`` or
            shorthand ``[duration, float('nan')]``. Also accepts
            ``[start, end]`` for absolute SUBSET requests.
        signals
            Channel selection — list of amp indices, ``SignalSpec.amp([...])``,
            or dict for multi-type.

        Returns
        -------
        SignalArrays
            Signal data as numpy arrays (channels x samples).
        """
        tr_spec = cast("TimeRangeSpec", time_range)
        sig_spec = cast("SignalSpec", signals)

        channel_meta = self.get_channel_metadata()
        sig_select = sig_spec.to_sig_select(channel_meta)

        def _sel(idxs: list[int]) -> common_pb2.HDSnapshotRequest.SelectedSignals | None:
            return common_pb2.HDSnapshotRequest.SelectedSignals(ntvChanIdxs=idxs) if idxs else None

        req = common_pb2.HDSnapshotRequest(
            streamGroupId=PRIMARY_CACHE_STREAM_GROUP_ID,
            timeRange=_time_range_floats(tr_spec),
            priSignals=_sel(sig_select.amp.tolist()),
            auxSignals=_sel(sig_select.gpio_ain.tolist()),
            dinSignals=_sel(sig_select.gpio_din.tolist()),
            doutSignals=_sel(sig_select.gpio_dout.tolist()),
        )

        response = get_hd_snapshot_proto(self._pool, self._pcache_address, req)
        return signal_arrays_from_hd_snapshot(response)

    def set_stream_state(self, request: StreamStateRequest) -> None:
        """Set data streaming state.

        Args:
            request: Stream state configuration specifying the desired mode

        Raises:
            RadiensError: If the stream state change fails
        """
        proto_request = stream_state_request_to_proto(request)
        response = set_stream_state_proto(self._pool, self._core_address, proto_request)

        if response.errMsg:
            raise RadiensError(f"Failed to set stream state: {response.errMsg}")

    def healthcheck(self, service: ServiceEndpoint = ServiceEndpoint.ALLEGO_LIFECYCLE) -> None:
        """Perform lightweight system health verification.

        Parameters
        ----------
        service
            Which gRPC service to probe (default ``ALLEGO_LIFECYCLE``).
            ``ALLEGO_CORE`` is also supported.

        Raises:
            RadiensError: If the health check fails or system is unresponsive
        """
        response = healthcheck_proto(self._pool, self._server_address(service), service=service)
        if response.errMsg:
            raise RadiensError(f"Health check failed: {response.errMsg}")

    def set_record_state(self, request: RecordStateRequest) -> None:
        """Set data recording state.

        Args:
            request: Record state configuration specifying the desired mode and optional port

        Raises:
            RadiensError: If the record state change fails
        """
        proto_request = record_state_request_to_proto(request)
        response = set_record_state_proto(self._pool, self._core_address, proto_request)
        if response.errMsg:
            raise RadiensError(f"Failed to set record state: {response.errMsg}")

    def restart(self, request: RestartRequest) -> None:
        """Restart the system with specified backbone mode.

        Args:
            request: Restart configuration specifying the desired backbone mode

        Raises:
            RadiensError: If the restart fails
        """
        proto_request = restart_request_to_proto(request)
        response = restart_proto(self._pool, self._core_address, proto_request)
        if response.errMsg:
            raise RadiensError(f"System restart failed: {response.errMsg}")

    def set_stim_params(self, params: dict[int, StimParams]) -> None:
        """Update stimulation parameters for specified channels.

        Args:
            params: Map from system channel indices to their new stimulation parameters.
                   Each StimParams defines waveform shape, timing, amplitudes, trigger
                   settings, pulse trains, and recovery parameters for that channel.

        Raises:
            RadiensError: On communication failure or server error

        Note:
            The underlying gRPC API sets parameters for one channel at a time.
            This method calls SetStimParams once per channel in the provided dictionary.

        Usage:
            Update stimulation parameters for multiple channels::

                # Define stimulation parameters
                stim_params = StimParams(
                    stim_shape=StimShape.BIPHASIC,
                    stim_polarity=StimPolarity.CATHODIC_FIRST,
                    first_phase_duration=100.0,  # 100 μs
                    second_phase_duration=100.0,  # 100 μs
                    interphase_delay=50.0,  # 50 μs
                    first_phase_amplitude=100.0,  # 100 μA
                    second_phase_amplitude=-100.0,  # 100 μA (opposite polarity)
                    baseline_voltage=0.0,
                    trigger_edge_or_level=StimTriggerEdge.EDGE,
                    trigger_high_or_low=StimTriggerLevel.HIGH,
                    enabled=True,
                    post_trigger_delay=1000.0,  # 1 ms = 1000 μs
                    pulse_or_train=StimPulseMode.SINGLE_PULSE,
                    number_of_stim_pulses=1,
                    pulse_train_period=20000.0,  # 20 ms = 50 Hz
                    refractory_period=10000.0,  # 10 ms
                    pre_stim_amp_settle=1000.0,  # 1 ms
                    post_stim_amp_settle=1000.0,  # 1 ms
                    maintain_amp_settle=False,
                    enable_amp_settle=True,
                    post_stim_charge_recov_on=1000.0,  # 1 ms
                    post_stim_charge_recov_off=0.0,
                    enable_charge_recovery=True,
                    trigger_source_is_keypress=False,
                    trigger_source_idx=0,
                    stim_sys_chan_idx=42,
                )

                # Update channels 42 and 43
                client.set_stim_params({
                    42: stim_params,
                    43: stim_params.model_copy(update={'first_phase_amplitude': 150.0}),
                })
        """

        if not params:
            msg = "params cannot be empty - must specify at least one channel"
            raise RadiensError(msg)

        # SetStimParams works on one channel at a time
        # Validate and process each channel
        for sys_chan_idx, stim_params in params.items():
            # Ensure the sys_chan_idx matches the one in the StimParams
            if stim_params.stim_sys_chan_idx != sys_chan_idx:
                # Create updated stim_params with correct sys_chan_idx
                updated_params = stim_params.model_copy(update={"stim_sys_chan_idx": sys_chan_idx})
            else:
                updated_params = stim_params

            proto_request = stim_params_to_proto(updated_params)
            response = set_stim_params_proto(self._pool, self._core_address, proto_request)

            # Check for errors in the response
            if response.errMsg:
                raise RadiensError(f"Set stimulation parameters failed for channel {sys_chan_idx}: {response.errMsg}")

    def set_trigger_state(self, request: SetTriggerStateRequest) -> None:
        """Set trigger channel state for stimulation control.

        Args:
            request: Configuration for trigger channel state

        Raises:
            RadiensError: On communication failure or server error
        """
        proto_request = set_trigger_state_request_to_proto(request)
        response = set_trigger_state_proto(self._pool, self._core_address, proto_request)
        if response.errMsg:
            raise RadiensError(f"Set trigger state failed: {response.errMsg}")

    def get_trigger_state(self) -> TriggerState:
        """Get current trigger channel configuration.

        Returns:
            Current trigger state showing enabled channels and ports

        Raises:
            RadiensError: On communication failure or server error
        """
        response = get_trigger_state_proto(self._pool, self._core_address)
        return trigger_state_from_proto(response)

    def manual_stim_trigger(self, request: ManualStimTriggerRequest) -> None:
        """Manually trigger stimulation.

        Args:
            request: Manual trigger configuration

        Raises:
            RadiensError: On communication failure or server error
        """
        proto_request = manual_stim_trigger_request_to_proto(request)
        response = manual_stim_trigger_proto(self._pool, self._core_address, proto_request)
        if response.errMsg:
            raise RadiensError(f"Manual stim trigger failed: {response.errMsg}")

    def manual_stim_trigger_toggle(self, request: ManualStimTriggerToggleRequest) -> None:
        """Toggle manual stimulation trigger.

        Args:
            request: Manual trigger toggle configuration

        Raises:
            RadiensError: On communication failure or server error
        """
        proto_request = manual_stim_trigger_toggle_request_to_proto(request)
        response = manual_stim_trigger_toggle_proto(self._pool, self._core_address, proto_request)
        if response.errMsg:
            raise RadiensError(f"Manual stim trigger toggle failed: {response.errMsg}")

    def set_stim_step(self, request: StimStepRequest) -> None:
        """Set stimulation step size.

        Args:
            request: Stimulation step configuration

        Raises:
            RadiensError: On communication failure or server error
        """
        proto_request = stim_step_request_to_proto(request)
        response = set_stim_step_proto(self._pool, self._core_address, proto_request)
        if response.errMsg:
            raise RadiensError(f"Set stim step failed: {response.errMsg}")

    def get_stim_step(self) -> StimStepState:
        """Get current stimulation step size setting.

        Returns:
            Current stimulation step configuration

        Raises:
            RadiensError: On communication failure or server error
        """
        response = get_stim_step_proto(self._pool, self._core_address)
        return stim_step_state_from_proto(response)

    def set_recording_config(self, config: RecordingConfig) -> None:
        """Set recording configuration settings.

        Configures output file paths, naming, and recording behavior.
        This affects where recorded data will be saved when recording is started.

        Args:
            config: Recording configuration specifying file paths, naming,
                   and recording type (continuous/spikes/both)

        Raises:
            RadiensError: If the recording config change fails

        Example:
            Configure recording to save continuous data with timestamps::

                config = RecordingConfig(
                    base_file_name="my_recording",
                    base_file_path="/path/to/data",
                    recording_type=DataRecordingType.CONTINUOUS,
                    time_stamp=True,
                )
                client.set_recording_config(config)
        """
        proto_request = recording_config_to_proto(config)
        response = set_recording_config_proto(self._pool, self._core_address, proto_request)
        if response.errMsg:
            raise RadiensError(f"Failed to set recording config: {response.errMsg}")

    def get_recording_config(self) -> RecordingConfig:
        """Get current recording configuration settings.

        Returns:
            Current recording configuration including file paths, naming,
            and recording type settings

        Raises:
            RadiensError: On communication failure or server error
        """
        response = get_recording_config_proto(self._pool, self._core_address)
        return recording_config_from_proto(response)



"""System status domain models and enums."""

from __future__ import annotations

from enum import IntEnum

from pydantic import BaseModel, ConfigDict

from radiens_core.models.ports import GPIOChannelCount, Port, PortChannelCount

# ===================================================================
# System status enums
# ===================================================================


class BackboneMode(IntEnum):
    """Hardware backbone mode. Values match proto ``allego.BackboneMode``."""

    SMARTBOX_PRO = 0
    SMARTBOX_SIM_GEN_SINE = 1  # low freq
    SMARTBOX_SIM_GEN_SPIKES = 2
    SMARTBOX_SIM_DATASOURCE = 3
    OPEN_EPHYS_USB3 = 4
    INTAN_RECORDING_CONTROLLER_1024 = 5
    SMARTBOX_CLASSIC = 6
    INTAN_RECORDING_CONTROLLER_512 = 7
    OPEN_EPHYS_USB2 = 8
    INTAN_USB2 = 9
    SMARTBOX_SIM_GEN_SINE_MAPPED = 10
    SMARTBOX_SIM_GEN_SINE_HIGH_FREQ = 11
    SMARTBOX_SIM_GEN_SINE_MULTI_BAND = 12
    SMARTBOX_PRO_SINAPS_256 = 13
    SMARTBOX_PRO_SINAPS_1024 = 14
    XDAQ_ONE_REC = 15
    XDAQ_ONE_STIM = 16
    XDAQ_CORE_REC = 17
    XDAQ_CORE_STIM = 18
    SMARTBOX_PRO_SINAPS_512 = 19
    XDAQ_GEN2_ONE_STIM = 20
    XDAQ_GEN2_CORE_STIM = 21
    XDAQ_GEN2_ONE_REC = 22
    XDAQ_GEN2_CORE_REC = 23
    SMARTBOX_SIM_GEN_LFP = 24


class StreamMode(IntEnum):
    """Data stream mode. Values match proto ``allego.StreamMode``."""

    OFF = 0  # S_OFF
    ON = 1  # S_ON


class RecordMode(IntEnum):
    """Data recording mode. Values match proto ``allego.RecordMode``."""

    OFF = 0  # R_OFF
    ON = 1  # R_ON


class DataRecordingType(IntEnum):
    """Data recording type. Values match proto ``allego.DataRecordingType``."""

    CONTINUOUS = 0  # continuous
    SPIKES = 1  # spikes
    CONTINUOUS_AND_SPIKES = 2  # continuousAndSpikes


# ===================================================================
# Status models
# ===================================================================


class RestartRequest(BaseModel):
    """Request to restart the system with a specific backbone mode."""

    model_config = ConfigDict(frozen=True)

    mode: BackboneMode


class RecordStateRequest(BaseModel):
    """Request to change recording state."""

    model_config = ConfigDict(frozen=True)

    mode: RecordMode
    port: Port | None = None  # If None, system-level record control


class StreamStateRequest(BaseModel):
    """Request to change streaming state."""

    model_config = ConfigDict(frozen=True)

    mode: StreamMode


class StreamingStatus(BaseModel):
    """Current streaming status."""

    model_config = ConfigDict(frozen=True)

    stream_mode: StreamMode
    primary_cache_t_range: list[float]
    hardware_memory_level: float


class RecordingStatus(BaseModel):
    """Current recording status."""

    model_config = ConfigDict(frozen=True)

    record_mode: RecordMode
    active_filename: str
    duration: float
    error: str
    record_mode_port_a: RecordMode | None = None
    record_mode_port_b: RecordMode | None = None
    record_mode_port_c: RecordMode | None = None
    record_mode_port_d: RecordMode | None = None
    record_mode_port_e: RecordMode | None = None
    record_mode_port_f: RecordMode | None = None
    record_mode_port_g: RecordMode | None = None
    record_mode_port_h: RecordMode | None = None


class RecordingConfig(BaseModel):
    """Recording configuration settings.

    Configures output file paths, naming, and recording behavior.
    Used as both request (for set_recording_config) and response (for get_recording_config).
    """

    model_config = ConfigDict(frozen=True)

    base_file_name: str
    """Base filename for recording output (without extension or timestamp)."""

    base_file_path: str
    """Base directory path for recording files."""

    datasource_idx: int = 0
    """Index number for the composed file name."""

    time_stamp: bool = True
    """True to include timestamp in the composed file name."""

    split_files_by_port: bool = False
    """True to create separate files for each port."""

    force_start_at_tstamp_0: bool = False
    """True to force recording to start at timestamp 0."""

    recording_type: DataRecordingType = DataRecordingType.CONTINUOUS
    """Type of data to record (continuous, spikes, or both)."""


class BackboneStatus(BaseModel):
    """Complete system backbone status."""

    model_config = ConfigDict(frozen=True)

    streaming: StreamingStatus
    recording: RecordingStatus
    ports: list[PortChannelCount]
    is_connected: bool
    gpio_channel_count: GPIOChannelCount



"""Generate the code reference pages and navigation."""

from pathlib import Path

import mkdocs_gen_files

nav = mkdocs_gen_files.Nav()

src_root = Path("src")

for path in sorted(src_root.rglob("*.py")):
    module_path = path.relative_to(src_root).with_suffix("")
    doc_path = path.relative_to(src_root).with_suffix(".md")
    full_doc_path = Path("reference", doc_path)

    parts = tuple(module_path.parts)

    if parts[-1] == "__init__":
        parts = parts[:-1]
        doc_path = doc_path.with_name("index.md")
        full_doc_path = full_doc_path.with_name("index.md")
    elif parts[-1] == "__main__":
        continue

    # Skip _generated and other internal folders if desired
    # We allow _transport, _converters, _services because they might have useful internal docs,
    # but usually we'd only want public stuff.
    # The filter in mkdocs.yml (filters: ["!^_[^_]"]) handles hiding private members.
    if "_generated" in parts:
        continue

    nav[parts] = doc_path.as_posix()

    with mkdocs_gen_files.open(full_doc_path, "w") as fd:
        ident = ".".join(parts)
        fd.write(f"# {ident}\n\n::: {ident}")


with mkdocs_gen_files.open("reference/SUMMARY.md", "w") as nav_file:
    nav_file.writelines(nav.build_literate_nav())



"""Radiens Python client for neuroscience data acquisition and analysis."""

from importlib.metadata import version

__version__ = version("radiens-core")

from radiens_core.allego_client import AllegoClient
from radiens_core.curate_client import CurateClient
from radiens_core.models.common import ServiceEndpoint
from radiens_core.videre_client import VidereClient

__all__ = [
    "AllegoClient",
    "CurateClient",
    "ServiceEndpoint",
    "VidereClient",
    "__version__",
]
